#ifndef CLIENT_H
#define CLIENT_H

#include "../engine/Wrapper.h"
#include "../engine/KeyManager.h"
#include "../engine/MouseManager.h"
#include "../engine/Clock.h"
#include "../engine/Shader.h"
#include "../engine/DirectRenderer.h"
#include "../math/Camera3D.h"
#include "../world/World.h"
#include "../world/IChunkProvider.h"
#include "rendering/ClientChunkProvider.h"
#include "rendering/ChunkRenderer.h"
#include "rendering/entity/EntityRenderer.h"

using namespace std;

class Client : public IClient
{
public:
    Client();
    virtual ~Client();
    
    void tick() override;
    void render3DShadowTick(float lag) override;
    void render3DTick(float lag) override;
    void render2DTick(float lag) override;
    
    void onKeyEvent(int key, int scancode, int action, int mods) override;
    void onMouseMove(double x, double y) override;
    void onMouseClick(int button, int action, int mods) override;
private:
    double transformMouse(double in);
    
    static const int KEY_LEFT = 0;
    static const int KEY_RIGHT = 1;
    static const int KEY_UP = 2;
    static const int KEY_DOWN = 3;
    static const int KEY_JUMP = 4;
    static const int KEY_SNEAK = 5;
    
    static const int KEY_TEST = 10;
    
    static const int MOUSE_LEFT = 0;
    
    double oldMouseX = 0;
    double oldMouseY = 0;
    double diffMouseX = 0;
    double diffMouseY = 0;
    bool mouseTrapped = true;
    const double MOUSE_BORDER = 1.0;
    
    Clock tps;
    Clock fps;
    
    KeyManager keyManager;
    MouseManager mouseManager;
    
    ChunkRenderer chunkRenderer;
    ClientChunkProvider chunkProvider;
    World world;
    
    Camera3D camera;
    Shader shader;
    DirectRenderer directRenderer;

    Vector3D position;
    float lengthAngle = 0;
    float widthAngle = 0;
    
    EntityRenderer entity;
};

#endif

